@@ -1,5 +1,4 @@
-#!/usr/bin/env python
-# -*- coding: UTF-8 -*-
+# -*- coding: utf-8 -*-
 """ File Input/Output Utility Methods """
 
 
@@ -9,7 +8,8 @@ from io import StringIO
 from io import open as io_open
 
 from csv import reader as csv_reader
-from typing import Generator
+from typing import List
+from typing import Optional
 
 from sys import platform
 from pathlib import Path
@@ -30,6 +30,21 @@ from yaml import load as yaml_load
 class FileIO(object):
     """ File Input/Output Utility Methods """
 
+    @staticmethod
+    def is_empty_folder(folder_name: str) -> bool:
+        """ Check if a Folder is Empty of Contents
+
+        Args:
+            folder_name (str): a fully qualified path to a folder
+
+        Returns:
+            bool: True if the folder has no contents
+        """
+        if FileIO.exists(folder_name):
+            if len(os.listdir(folder_name)):
+                return False
+        return True
+
     @staticmethod
     def local_directory() -> str:
         """ Retrieve a Platform Specific Local Directory
@@ -40,11 +55,11 @@ class FileIO(object):
         Returns:
             str: absolute Path to a Local Directory
         """
-        if platform == "linux" or platform == "linux2":
+        if platform == 'linux' or platform == 'linux2':
             return os.environ['HOME']
-        elif platform == "darwin":
+        elif platform == 'darwin':
             return os.environ['HOME']
-        elif platform == "win32":
+        elif platform == 'win32':
             return os.environ['APPDATA']
         raise NotImplementedError(platform)
 
@@ -114,13 +129,13 @@ class FileIO(object):
         if _type not in [dict, list, str]:
             raise NotImplementedError(_type)
 
-        path = FileIO.join(FileIO.local_directory(), "temp.json")
+        path = FileIO.join(FileIO.local_directory(), 'temp.json')
         if _type == str:
             FileIO.write_string(data, path)
         else:
             FileIO.write_json(data, path)
 
-        print(f"Wrote File to {path}")
+        print(f'Wrote File to {path}')
         return path
 
     @staticmethod
@@ -195,7 +210,7 @@ class FileIO(object):
         """
         file_path = os.path.join(os.environ['DESKTOP'], 'temp.json')
         FileIO.write_json(data, file_path=file_path)
-        raise ValueError(f"Wrote Temp File: {file_path}")
+        raise ValueError(f'Wrote Temp File: {file_path}')
 
     @staticmethod
     def write_image(data: object,
@@ -218,7 +233,7 @@ class FileIO(object):
             data (object): the JSON object
             file_path (str): the absolute and qualified output file path
         """
-        with io_open(file_path, "w") as json_file:
+        with io_open(file_path, 'w') as json_file:
             json_dump = json_dumps(data,
                                    indent=4,
                                    sort_keys=False,
@@ -241,7 +256,7 @@ class FileIO(object):
 
     @staticmethod
     def read_string(file_path: str,
-                    encoding: str = "utf-8",
+                    encoding: str = 'utf-8',
                     replace_newlines: bool = False) -> str:
         """ Read String from File
 
@@ -260,7 +275,7 @@ class FileIO(object):
     @staticmethod
     def write_string(input_text: str,
                      file_path: str,
-                     file_encoding: str = "utf-8") -> None:
+                     file_encoding: str = 'utf-8') -> None:
         """ Write String to File
 
         Args:
@@ -275,7 +290,7 @@ class FileIO(object):
             raise ValueError
 
         target = codecs_open(file_path,
-                             mode="w",
+                             mode='w',
                              encoding=file_encoding)
 
         target.write(input_text)
@@ -284,7 +299,7 @@ class FileIO(object):
     @staticmethod
     def write_lines(lines: list,
                     file_path: str,
-                    file_encoding: str = "utf-8") -> None:
+                    file_encoding: str = 'utf-8') -> None:
         """ Write Lines to File
 
         Args:
@@ -340,11 +355,11 @@ class FileIO(object):
             try:
                 return yaml_load(stream, Loader)
             except YAMLError:
-                raise ValueError(f"Invalid YAML File: {file_path}")
+                raise ValueError(f'Invalid YAML File: {file_path}')
 
     @staticmethod
     def read_lines(file_path: str,
-                   file_encoding: str = "utf-8") -> list:
+                   file_encoding: str = 'utf-8') -> list:
         """ Read Lines of a File into a List
 
         Args:
@@ -355,16 +370,16 @@ class FileIO(object):
             list: each line in the file as a list
         """
         target = codecs_open(file_path,
-                             mode="r",
+                             mode='r',
                              encoding=file_encoding)
-        lines = [x.replace("\n", "").strip() for x in target.readlines() if x]
+        lines = [x.replace('\n', '').strip() for x in target.readlines() if x]
         target.close()
 
         return lines
 
     @staticmethod
     def yield_lines(file_path: str,
-                    file_encoding: str = "utf-8") -> list:
+                    file_encoding: str = 'utf-8') -> list:
         """Read Lines of a File using a yield Generator
 
         Args:
@@ -385,7 +400,7 @@ class FileIO(object):
             Notes:
                 each output line is a list with one element that requires independent delimitation
         """
-        with codecs_open(file_path, mode="r", encoding=file_encoding) as f:
+        with codecs_open(file_path, mode='r', encoding=file_encoding) as f:
             for line in csv_reader(f):
                 yield line
 
@@ -432,19 +447,61 @@ class FileIO(object):
         return files
 
     @staticmethod
-    def load_files(directory: str,
-                   extension: str,
-                   limit: int = None,
-                   blacklist: list = None,
-                   recursive: bool = False) -> list:
-        """ Load Files from a directory
+    def count_files(directory: str,
+                    extension: str,
+                    recursive: bool = False) -> Optional[int]:
+        """ Count Files in a directory
 
         Args:
             directory (str): the absolute and qualified directory path
             extension (str): the file extension to load
             recursive (bool, optional): loads files recursively if set to True. Defaults to False.
-            limit (int, optional): The number of files to load. Defaults to None.
-            blacklist (list, optional): an list of File Names to exclude. Defaults to None.
+
+        Returns:
+            Optional[int]: the total files (0..*)
+                if the directory does not exist, return None
+        """
+        if not FileIO.exists(directory):
+            return None
+
+        return len(FileIO.load_files(directory=directory,
+                                     extension=extension,
+                                     limit=None,
+                                     blacklist=None,
+                                     recursive=recursive))
+
+    @staticmethod
+    def list_directory(directory: str,
+                       recursive: bool = True) -> list:
+        """ Load Files from a directory
+
+        Args:
+            directory (str): the absolute and qualified directory path
+            recursive (bool, optional): loads files recursively if set to True. Defaults to True.
+
+        Returns:
+            list: list of files
+        """
+        return FileIO.load_files(directory=directory,
+                                 extension=None,
+                                 limit=None,
+                                 blacklist=None,
+                                 recursive=recursive)
+
+    @staticmethod
+    def load_files(directory: str,
+                   extension: Optional[str] = None,
+                   limit: Optional[int] = None,
+                   blacklist: Optional[List[str]] = None,
+                   recursive: bool = False) -> list:
+        """ Load Files from a directory
+
+        Args:
+            directory (str): the absolute and qualified directory path
+            extension (Optional[str], optional): the file extension to load. Defaults to None.
+            limit (Optional[int], optional): The number of files to load. Defaults to None.
+            blacklist (Optional[List[str]], optional): an list of File Names to exclude. Defaults to None.
+            recursive (bool, optional): loads files recursively if set to True. Defaults to False.
 
         Returns:
             list: list of files
@@ -452,20 +509,29 @@ class FileIO(object):
         def non_recursive_loader() -> list:
             files = os.listdir(directory)
 
-            files = [f for f in files if f.endswith(extension)]
+            if extension:
+                files = [f for f in files if f.endswith(extension)]
+
             if limit and len(files) >= limit:
                 files = files[:limit]
 
-            files = [os.path.normpath(os.path.join(directory, f))
-                     for f in files]
+            files = [
+                os.path.normpath(os.path.join(directory, f))
+                for f in files
+            ]
 
             return files
 
         def recursive_loader() -> list:
             results = []
 
-            for dirpath, dirnames, files in os.walk(directory):
-                files = [x for x in files if x.lower().endswith(extension)]
+            for dirpath, _, files in os.walk(directory):
+
+                if extension:
+                    files = [
+                        x for x in files
+                        if x.lower().endswith(extension)
+                    ]
 
                 for name in files:
 
@@ -516,7 +582,7 @@ class FileIO(object):
         exclude_sub = []
         for sub_folder in exclude:
             if os.path.sep not in sub_folder:
-                exclude_sub.append(f"{os.path.sep}{sub_folder}{os.path.sep}")
+                exclude_sub.append(f'{os.path.sep}{sub_folder}{os.path.sep}')
             else:
                 exclude_sub.append(sub_folder)
 
@@ -573,7 +639,7 @@ class FileIO(object):
         exclude_sub = []
         for sub_folder in exclude:
             if os.path.sep not in sub_folder:
-                exclude_sub.append(f"{os.path.sep}{sub_folder}{os.path.sep}")
+                exclude_sub.append(f'{os.path.sep}{sub_folder}{os.path.sep}')
             else:
                 exclude_sub.append(sub_folder)
 
@@ -632,7 +698,7 @@ class FileIO(object):
         try:
             return yaml_load(StringIO(file_data), Loader)
         except YAMLError:
-            raise ValueError("Invalid YAML Data")
+            raise ValueError('Invalid YAML Data')
 
     @staticmethod
     def parse_json(file_data: str) -> dict:
@@ -657,4 +723,4 @@ class FileIO(object):
             return json_loads(file_data)
 
         except JSONDecodeError:
-            raise ValueError("Invalid JSON Data")
+            raise ValueError('Invalid JSON Data')
