@@ -1,5 +1,4 @@
-#!/usr/bin/env python
-# -*- coding: UTF-8 -*-
+# -*- coding: utf-8 -*-
 """ Text Utility Methods: Common Functions without Special Libraries """
 
 
@@ -32,6 +31,85 @@ STOPWORDS = [
 class TextUtils(object):
     """ Text Utility Methods: Common Functions without Special Libraries """
 
+    @staticmethod
+    def chunk_list(input_lines: List[str],
+                   max_chunk_size: int) -> List[str]:
+        """ Rechunk a List of Strings by Size
+
+        Sample Input (5 entries):
+            [
+                input sentence 1,
+                input sentence 2,
+                input sentence 3,
+                input sentence 4,
+                input sentence 5,
+            ]
+
+        Sample Call
+            get_chunks(input_lines, 50)
+
+        Sample Output (2 entries)
+            [
+                input sentence 1 input sentence 2 input sentence 3,
+                input sentence 4 input sentence 5,
+            ]
+
+        Args:
+            input_lines (List[str]): a list of string inputs
+            max_chunk_size (int): the desired chunk size
+
+        Returns:
+            List[str]: a list of string outputs
+        """
+
+        chunks = []
+        buffer = []
+        running_total = 0
+
+        for i in range(len(input_lines)):
+
+            buffer.append(input_lines[i])
+            running_total += len(input_lines[i])
+
+            if running_total > max_chunk_size:
+                chunks.append('  '.join(buffer))
+                buffer = []
+                running_total = 0
+
+        if len(buffer):
+            chunks.append('  '.join(buffer))
+
+        return chunks
+
+    @staticmethod
+    def is_equal(text_1: str,
+                 text_2: str,
+                 case_sensitive: bool = False) -> bool:
+        """ Compare Two Complex Strings for Hash Equality
+
+        Args:
+            text_1 (str): any input text of any length or format
+            text_2 (str): any input text of any length or format
+            case_sensitive (bool, optional): Will ignore case qualifications if True. Defaults to False.
+
+        Returns:
+            bool: True if the strings are equal
+        """
+
+        # try a simple equality op first
+        if hash(text_1) == hash(text_2):
+            return True
+
+        def cleanse(input_text: str) -> str:
+            if not case_sensitive:
+                input_text = input_text.lower()
+            for ch in ['\n', '\r', '\t', ' ']:
+                while ch in input_text:
+                    input_text = input_text.replace(ch, '').strip()
+            return input_text.strip()
+
+        return hash(cleanse(text_1)) == hash(cleanse(text_2))
+
     @staticmethod
     def remove_duplicated_phrases(text_1: str,
                                   text_2: str) -> str:
@@ -55,10 +133,11 @@ class TextUtils(object):
         return text_1
 
     @staticmethod
+    # TODO: 20230116; this seems like an unexpected usage of itertools.product(...)
     def cartesian(matches: list) -> list:
         """
         Purpose:
-            Find a list of candidate token sequences within the normalized_input_text
+            Perform a Cartesian Join
         Time per Call:
             0.5ms < x 1.0ms
         :param matches:
@@ -74,6 +153,158 @@ class TextUtils(object):
 
         return cartesian
 
+    @staticmethod
+    def cartesian_bigrams(matches: list) -> list:
+        """
+        Purpose:
+            Perform a Cartesian Join to create Bigram Products
+        :param matches:
+            a list of 1..* tokens forming a match pattern
+            Sample Input:
+                [
+                    'Light', 
+                    'Lighting', 
+                    'Purple', 
+                    'People', 
+                    'Person',
+                    'Laser', 
+                    'Shoe', 
+                    'Face', 
+                    'Head', 
+                    'Graduation'
+                ]
+        :return:
+            a list of possible token sequences
+            Sample Output:
+                [
+                    ['Face', 'Graduation'], 
+                    ['Face', 'Head'], 
+                    ['Face', 'Laser'], 
+                    ['Face', 'Light'], 
+                    ['Face', 'Lighting'], 
+                    ['Face', 'People'], 
+                    ['Face', 'Person'], 
+                    ['Face', 'Purple'], 
+                    ['Face', 'Shoe'],
+                    ['Graduation', 'Head'], 
+                    ['Graduation', 'Laser'], 
+                    ['Graduation', 'Light'], 
+                    ['Graduation', 'Lighting'], 
+                    ['Graduation', 'People'], 
+                    ['Graduation', 'Person'], 
+                    ['Graduation', 'Purple'], 
+                    ['Graduation', 'Shoe'], 
+                    ['Head', 'Laser'], 
+                    ['Head', 'Light'], 
+                    ['Head', 'Lighting'], 
+                    ['Head', 'People'], 
+                    ['Head', 'Person'], 
+                    ['Head', 'Purple'], 
+                    ['Head', 'Shoe'], 
+                    ['Laser', 'Light'], 
+                    ['Laser', 'Lighting'], 
+                    ['Laser', 'People'], 
+                    ['Laser', 'Person'], 
+                    ['Laser', 'Purple'], 
+                    ['Laser', 'Shoe'], 
+                    ['Light', 'Lighting'], 
+                    ['Light', 'People'], 
+                    ['Light', 'Person'], 
+                    ['Light', 'Purple'], 
+                    ['Light', 'Shoe'], 
+                    ['Lighting', 'People'],
+                    ['Lighting', 'Person'],
+                    ['Lighting', 'Purple'], 
+                    ['Lighting', 'Shoe'], 
+                    ['People', 'Person'], 
+                    ['People', 'Purple'],
+                    ['People', 'Shoe'], 
+                    ['Person', 'Purple'],
+                    ['Person', 'Shoe'], z
+                    ['Purple', 'Shoe']
+                ]
+        """
+        product = TextUtils.cartesian([matches, matches])
+
+        product = [x for x in product if x[0] != x[1]]
+
+        def sort_key(s): return (-len(s), s)
+
+        product = sorted(set(
+            [
+                ','.join(sorted(x, key=sort_key)) for x in product
+            ]
+        ))
+
+        product = [x.split(',') for x in product]
+
+        return product
+
+    @staticmethod
+    def cartesian_trigrams(matches: list) -> list:
+        """
+        Purpose:
+            Perform a Cartesian Join to create Bigram Products
+
+        :param matches:
+            a list of 1..* tokens forming a match pattern
+
+        Returns:
+            list: a list of trigrams
+        """
+
+        product = TextUtils.cartesian([matches, matches, matches])
+
+        product = [
+            x for x in product if
+            x[0] != x[1] and
+            x[0] != x[2] and
+            x[1] != x[2]
+        ]
+
+        def sort_key(s): return (-len(s), s)
+
+        product = sorted(set(
+            [
+                ','.join(sorted(x, key=sort_key)) for x in product
+            ]
+        ))
+
+        product = [x.split(',') for x in product]
+
+        return product
+
+    @staticmethod
+    def cartesian_quadgrams(matches: list) -> list:
+        """
+        Purpose:
+            Perform a Cartesian Join to create Bigram Products
+
+        :param matches:
+            a list of 1..* tokens forming a match pattern
+
+        Returns:
+            list: a list of trigrams
+        """
+
+        product = TextUtils.cartesian([matches, matches, matches, matches])
+
+        product = [
+            x for x in product if x[0] != x[1] and x[0] != x[2] and x[0] != x[3] and x[1] != x[2] and x[1] != x[3] and x[2] != x[3]
+        ]
+
+        def sort_key(s): return (-len(s), s)
+
+        product = sorted(set(
+            [
+                ','.join(sorted(x, key=sort_key)) for x in product
+            ]
+        ))
+
+        product = [x.split(',') for x in product]
+
+        return product
+
     @staticmethod
     def sliding_window(tokens: List[str],
                        window_size: int) -> Optional[List[str]]:
@@ -159,14 +390,14 @@ class TextUtils(object):
             for item_2 in [' '.join(x) for x in t2]:
 
                 if item_1 == item_2:
-                    return {100: {"tokens_1": item_1, "tokens_2": item_2}}
+                    return {100: {'tokens_1': item_1, 'tokens_2': item_2}}
 
                 score = TextUtils.jaccard_similarity(item_1, item_2)
                 if debug:
-                    print(f"({item_1}) vs. ({item_2}) = {score}")
+                    print(f'({item_1}) vs. ({item_2}) = {score}')
 
                 if score >= score_threshold:
-                    d_results[score] = {"tokens_1": item_1, "tokens_2": item_2}
+                    d_results[score] = {'tokens_1': item_1, 'tokens_2': item_2}
 
         return d_results
 
@@ -257,7 +488,7 @@ class TextUtils(object):
             return True
 
         cat = category(char)
-        if cat.startswith("P"):
+        if cat.startswith('P'):
             return True
 
         return False
@@ -281,7 +512,11 @@ class TextUtils(object):
             if not TextUtils.is_punctuation(ch):
                 buffer.append(ch)
 
-        return ''.join(buffer)
+        result = ''.join(buffer)
+        while '  ' in result:
+            result = result.replace('  ', ' ')
+
+        return result.strip()
 
     @ staticmethod
     def find_subsumed_tokens(tokens: list) -> list:
@@ -308,15 +543,15 @@ class TextUtils(object):
         """
 
         def exists(item_1: str, item_2: str) -> bool:
-            token_lr = f" {item_1} "
+            token_lr = f' {item_1} '
             if token_lr in item_2:
                 return True
 
-            token_l = f" {item_1}"
+            token_l = f' {item_1}'
             if item_2.endswith(token_l):
                 return True
 
-            token_r = f"{item_1} "
+            token_r = f'{item_1} '
             if item_2.startswith(token_r):
                 return True
 
@@ -593,8 +828,8 @@ class TextUtils(object):
 
             def an() -> str:
                 if tokens[i].isupper():
-                    return "An"
-                return "an"
+                    return 'An'
+                return 'an'
 
             if use_an():
                 results.append(an())
@@ -656,7 +891,7 @@ class TextUtils(object):
                 'The quick brown fox jumped over the lazy dog'
         """
         def case(value: str) -> str:
-            return f"{value[:1].upper()}{value[1:]}"
+            return f'{value[:1].upper()}{value[1:]}'
 
         if ' ' not in input_text:
             return case(input_text)
@@ -691,7 +926,7 @@ class TextUtils(object):
                 'The Quick Brown Fox Jumped Over the Lazy Dog'
         """
         def case(value: str) -> str:
-            return f"{value[:1].upper()}{value[1:]}"
+            return f'{value[:1].upper()}{value[1:]}'
 
         def conditional_case(value: str) -> str:
             if value.lower().strip() in STOPWORDS:
